(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2005-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: pio.ml,v 1.24 2014/03/04 18:41:51 fclement Exp $ *)

open Communication;;

(* {3 General send communication functions on a buffer} *)

let send_communication = Write_communication.write_communication;;

let send_values ob vals = send_communication ob (Phrase vals);;

let send_one_value ob v = send_values ob [| v |];;

(* {6 Specialized send communication functions on a buffer} *)

let gen_send of_native ob t =
  Write_typed_value.write_typed_value ob (of_native t)
;;

let send_string = gen_send Typed_value.of_string
and send_int = gen_send Typed_value.of_int
and send_int_couple = gen_send Typed_value.of_int_couple
and send_int_triple = gen_send Typed_value.of_int_triple
and send_float = gen_send Typed_value.of_float
and send_float_couple = gen_send Typed_value.of_float_couple
and send_float_triple = gen_send Typed_value.of_float_triple
and send_string_vector = gen_send Typed_value.of_string_vector
and send_int_vector = gen_send Typed_value.of_int_vector
and send_int_couple_vector = gen_send Typed_value.of_int_couple_vector
and send_int_triple_vector = gen_send Typed_value.of_int_triple_vector
and send_float_vector = gen_send Typed_value.of_float_vector
and send_float_couple_vector = gen_send Typed_value.of_float_couple_vector
and send_float_triple_vector = gen_send Typed_value.of_float_triple_vector
and send_c_int_matrix = gen_send Typed_value.of_c_int_matrix
and send_fortran_int_matrix = gen_send Typed_value.of_fortran_int_matrix
and send_c_float_matrix = gen_send Typed_value.of_c_float_matrix
and send_fortran_float_matrix = gen_send Typed_value.of_fortran_float_matrix
;;

(* General receive communication from a scanning buffer *)

let receive_communication = Read_communication.read_communication;;

let receive_values ib =
  match receive_communication ib with
  | Phrase vals -> vals
  | Task _ | Result _ | Error _ | Service _ ->
    failwith "Pio.receive_values: communication is not a phrase"
;;

let receive_one_value ib =
  let vals = receive_values ib in
  if Array.length vals = 1 then vals.(0) else
  failwith
    "Pio.receive_one_value: \
     communication phrase has more than one value"
;;

(* Specialized receive communications from a buffer *)

let gen_receive to_native ib =
  try to_native (Read_typed_value.read_typed_value ib) with
  | End_of_file ->
    prerr_endline "Pio: premature end of communication";
    exit 2
  | x ->
    prerr_endline
      (Printf.sprintf "Pio: spurious exception: %s" (Printexc.to_string x));
    exit 2
;;

let receive_string = gen_receive Typed_value.to_string
and receive_int = gen_receive Typed_value.to_int
and receive_int_couple = gen_receive Typed_value.to_int_couple
and receive_int_triple = gen_receive Typed_value.to_int_triple
and receive_float = gen_receive Typed_value.to_float
and receive_float_couple = gen_receive Typed_value.to_float_couple
and receive_float_triple = gen_receive Typed_value.to_float_triple
and receive_string_vector = gen_receive Typed_value.to_string_vector
and receive_int_vector = gen_receive Typed_value.to_int_vector
and receive_int_couple_vector = gen_receive Typed_value.to_int_couple_vector
and receive_int_triple_vector = gen_receive Typed_value.to_int_triple_vector
and receive_float_vector = gen_receive Typed_value.to_float_vector
and receive_float_couple_vector =
  gen_receive Typed_value.to_float_couple_vector
and receive_float_triple_vector =
  gen_receive Typed_value.to_float_triple_vector
and receive_c_int_matrix = gen_receive Typed_value.to_c_int_matrix
and receive_fortran_int_matrix = gen_receive Typed_value.to_fortran_int_matrix
and receive_c_float_matrix = gen_receive Typed_value.to_c_float_matrix
and receive_fortran_float_matrix =
  gen_receive Typed_value.to_fortran_float_matrix
;;

(* Foreman/Worker. *)

module Cmd_opt = Ucommand_option;;

type pf = Udebug.printing_facilities;;
type spf = Udebug.specialized_printing_facilities;;

type 'a injector = 'a -> values;;
type 'a projector = values -> 'a;;

type ('a, 'b) task_wrapper =
  {
   fname : Upervasives.function_name;
   tname : task_name;
   spf : spf;
   inj : 'a injector;
   proj : 'b projector;
  }
;;

type ('a, 'b) task = spf -> 'a -> 'b;;

type welcome_action = Upervasives.gen_message;;
type farewell_action = Upervasives.gen_message;;

type task_computation =
  pf -> Uunix.process_io_buffers -> values -> unit
;;

type task_table = (task_name * task_computation) Ulist1.t;;

exception Unbound_communication of t;;
exception Unbound_service_communication of (service_name * values);;
exception Unbound_task of task_name;;
exception Uninitialized_worker of task_name;;

let task_prefix = Printf.sprintf "Task %S:";;

let make_task_wrapper tname mk_inj mk_proj pf fname =
  let spf =
    Udebug.prefix_messages
      (task_prefix tname)
      (Udebug.set_function_name fname pf) in
  {
   fname = fname;
   tname = tname;
   spf = spf;
   inj = mk_inj spf;
   proj = mk_proj spf;
  }
;;

let gen_make_pf sname =
  let purpose = Udebug.User in
  let key_prefixes = Udebug.make_key_prefixes sname in
  Udebug.make_printing_facilities_with_static_tag purpose sname key_prefixes
;;

let foreman_make_printing_facilities () = gen_make_pf "Foreman"
and worker_make_printing_facilities () = gen_make_pf "Worker"
;;

let worker_test_wrap_task piob tname inj input =
  let { Uunix.output_buffer = ob; flush_output = flush; _; } = piob in
  let input_values = inj input in
  send_communication ob (Task (tname, input_values));
  flush ();
;;

let foreman_wrap_task cmd piob_ref tw input =
  let { fname = fname; tname = tname;
        spf = { Udebug.spf_report = report; _; } as spf;
        inj = inj; proj = proj; } = tw in
  let ({ Uunix.input_buffer = ib; _; } as piob) =
    Uunix.spawn spf cmd piob_ref in
  report "started";
  worker_test_wrap_task piob tname inj input;
  report "task sent";
  let output =
    match receive_communication ib with
    | Result output_values ->
        report "result received";
        proj output_values
    | Error exns ->
        let sl = Array.to_list (Array.map Typed_value.to_string exns) in
        let msg = String.concat "" sl in
        Upervasives.failwith fname msg
    | Service s -> raise (Unbound_service_communication s)
    | Phrase _ | Task _ as com -> raise (Unbound_communication com) in
  report "finished";
  output
;;

let worker_wrap_task piob tw task input_values =
  let { spf = { Udebug.spf_report = report; _; } as spf;
        inj = inj; proj = proj; _; } = tw in
  let { Uunix.output_buffer = ob; flush_output = flush; _; } = piob in
  report "started";
  let input = proj input_values in
  let output = task spf input in
  report "result computed";
  let output_values = inj output in
  send_communication ob (Result output_values);
  flush ();
  report "result sent back";
  report "finished";
;;

let spread_out_task tname tbl pf piob vals =
  let compute =
    try Ulist1.assoc tname tbl with
    | Not_found -> raise (Unbound_task tname) in
  compute pf piob vals
;;

let spread_out_communication tbl pf piob =
  let { Udebug.report; _; } = pf in
  let { Uunix.input_buffer = ib; _; } = piob in
  match receive_communication ib with
  | Task (tname, vals) ->
      report (Upervasives.prefix (task_prefix tname) "received");
      spread_out_task tname tbl pf piob vals
  | Service s -> raise (Unbound_service_communication s)
  | Phrase _ | Result _ | Error _ as com -> raise (Unbound_communication com)
;;

let rec loop tbl pf piob =
  spread_out_communication tbl pf piob;
  loop tbl pf piob;
;;

let init_required tname =
  Upervasives.prefix (task_prefix tname) "requires initialization"
;;

let spurious_exn exn =
  Printf.sprintf "Spurious exception: %S" (Printexc.to_string exn)
;;

let worker_main hi_action bye_action tbl =
  let ({ Udebug.report = report; error = error; _; } as pf) =
    worker_make_printing_facilities () in
  Cmd_opt.parse ();
  let failwith msg = error "" msg; exit 1 in
  report (hi_action ());
  let piob = Uunix.std_process_io_buffers in
  begin
    try loop tbl pf piob with
    | Exit -> ()
    | Wrong_communication msg -> failwith msg
    | Uninitialized_worker tname -> failwith (init_required tname)
    | Unbound_task _
    | Unbound_communication _
    | Unbound_service_communication _ as exn ->
        failwith (Printexc.to_string exn)
    | exn -> failwith (spurious_exn exn)
  end;
  report (bye_action ());
;;
